#
#	pandocFilter.py
#
#	Script to convert a oneM2M Markdown file to pandoc input format
#
#	(c) 2023 by Andreas Kraft, Miguel Angel Reina Ortega
#	License: BSD 3-Clause License. See the LICENSE file for further details.
#

import argparse, os, re, sys
from rich import print, inspect
from rich.progress import Progress, TextColumn, TimeElapsedColumn

def readMDFile(progress:Progress, document:str) -> list[str]:
	"""	Read the markdown file and return a list of lines.
	"""	
	_taskID = progress.add_task('[blue]Reading document', total=0)

	# Check if file exists
	if not os.path.exists(document):
		print(f'File {document} does not exist')
		exit(1)

	# Read the file
	with open(document, 'r', encoding='utf-8', errors = 'replace') as f:
		progress.stop_task(_taskID)
		return f.readlines()
	

def writeMDFile(progress:Progress, mdLines:list[str], document:str, outDirectory:str) -> None:
	"""	Write the markdown file.

	"""
	_taskID = progress.add_task('[blue]Writing document', total=0)

	# Write the file
	with open(f'{outDirectory}/{os.path.basename(document)}', 'w', encoding='utf-8', errors = 'replace') as f:
		f.writelines(mdLines)
	progress.stop_task(_taskID)


_inCodeBlock = False
def checkInCodeBlock(line:str) -> bool:
	global _inCodeBlock

	# Check if we are in a code block
	if line.strip().startswith('```'):
		_inCodeBlock = not _inCodeBlock
	return _inCodeBlock


def correctTOC(progress:Progress, mdLines:list[str], tocSection:str = 'Contents') -> list[str]:
	"""	Correct the TOC to be compatible with pandoc.
	"""
	_taskID = progress.add_task('[blue]Correcting TOC', total=0)

	_contents = f'# {tocSection}\n'
	tocregex = re.compile('^(.*\[.*\])')
	
	_lines:list[str] = []
	_inTOC = False
	for line in mdLines:
		# find TOC section first
		if line == _contents:
			_inTOC = True
			_lines.append(line)
			continue
		if _inTOC:
			if line.startswith('#'):	# End of TOC?
				_inTOC = False
				_lines.append(f'\n')
				_lines.append(line)
				continue
			matches = re.findall(tocregex, line)	# Replace entry
			if matches:
				_lines.append(f'{matches[0]}  \n')
				continue
			else:
				_lines.append(line)
		else:
			_lines.append(line)

	progress.stop_task(_taskID)
	return _lines


def replaceTableCaptions(progress:Progress, mdLines:list[str]) -> list[str]:
	"""	Replace table captions with a pandoc table caption.
	"""
	_taskID = progress.add_task('[blue]Replacing table captions', total=0)
	# progress.update()
	tableregex = re.compile('^\*\*(Table .*)\*\*')
	
	_lines:list[str] = []
	for line in mdLines:
		matches = re.findall(tableregex, line)
		if matches:
			# move the caption to the beginning of the table and add a "Table:" prefix
			_idx = len(_lines) - 1
			while _idx >= 0 and _lines[_idx].startswith('|'):
				_idx -= 1
			#if _idx > 0:
			#	_lines.insert(_idx+1, f'Table: {matches[0]}\n')
			_lines.insert(_idx+1, f'Table: {matches[0]}\n')
		else:
			_lines.append(line)

	#print(_lines)
	progress.stop_task(_taskID)
	return _lines


def replaceFigureCaptions(progress:Progress, mdLines:list[str]) -> list[str]:
	"""	Replace figure captions with a pandoc figure caption.
	"""
	_taskID = progress.add_task('[blue]Replacing figure captions', total=0)
	# progress.update()
	figureregex = re.compile('^\*\*(Figure .*)\*\*')
	nocaptionfigureregex = re.compile('!\[~~CAPTION~~\].*')
	
	_lines:list[str] = []
	for line in mdLines:
		matches = re.findall(figureregex, line)
		if matches:
			# Replace the previous figure markdown name with the captions
			_idx = len(_lines) - 1
			while _idx >= 0 and _lines[_idx].count('![') == 0:
				_idx -= 1
			if _idx > 0:
				_lines[_idx] = re.sub(r'!.*]', f'![{matches[0]}]', _lines[_idx])
		elif re.findall(nocaptionfigureregex, line):
			_lines.append(re.sub(r'^.*?]', f'![]', line))
		else:
			_lines.append(line)

	progress.stop_task(_taskID)
	return _lines


def replaceFiguresPathSvgToPng(progress: Progress, mdLines: list[str]) -> list[str]:
	"""	Replace figure extensions from svg to png.
	"""
	_taskID = progress.add_task('[blue]Replacing figure captions', total=0)
	# progress.update()
	figurePathRegex = re.compile('media\/.*\.svg')

	_lines: list[str] = []
	for line in mdLines:
		matches = re.findall(figurePathRegex, line)
		if matches:
			# Replace figure path to png
			_lines.append(re.sub(r'\.svg', f'.png', line))
		else:
			_lines.append(line)

	progress.stop_task(_taskID)
	return _lines


def replaceLineBreaks(progress: Progress, mdLines: list[str]) -> list[str]:
	"""	Replace <br /> linebreaks by pandoc escaped_line_breaks extension \(newline).
	"""
	_taskID = progress.add_task('[blue]Replacing linebreaks', total=0)
	# progress.update()
	linebreaksregex = re.compile('<\s*br\s*/?\s*>')

	_lines: list[str] = []
	_inCodeBlock = False
	for line in mdLines:
		# Check if we are in a code block
		if line.strip().startswith('```'):
			_inCodeBlock = not _inCodeBlock
		if _inCodeBlock:
			_lines.append(line)
			continue

		# Replace linebreaks
		matches = re.findall(linebreaksregex, line)
		if matches:
			# Replace the linebreak with "\(newline)"
			_lines.append(re.sub(linebreaksregex, f'\\\n', line))
		else:
			_lines.append(line)

	progress.stop_task(_taskID)
	return _lines


def correctTableSeparators(progress: Progress, mdLines: list[str]) -> list[str]:
	"""	Replace all table header separators with very simple ones. This is
		needed because pandoc calculates the column width based on the separators.
		And that is not always wanted
	"""
	_taskID = progress.add_task('[blue]Correcting table separators', total=0)
	tableSeparatorRegex = re.compile('^(\s*)(\|[-\|]*\|)\s*$')

	_lines: list[str] = []
	for line in mdLines:
		if checkInCodeBlock(line):
			_lines.append(line)
			continue

		match = re.match(tableSeparatorRegex, line)
		if match:
			# Replace the separators with a simple one
			_lines.append(f'{match.group(1)}{re.sub("-+", "-", match.group(2))}\n')
		else:
			_lines.append(line)

	progress.stop_task(_taskID)
	return _lines


def process(document:str, outDirectory:str) -> None:
	with Progress(TextColumn('{task.description}'),  TimeElapsedColumn()) as progress:
		mdLines = readMDFile(progress, document)
		mdLines = correctTOC(progress, mdLines)
		mdLines = replaceTableCaptions(progress, mdLines)
		mdLines = replaceFigureCaptions(progress, mdLines)
		mdLines = replaceFiguresPathSvgToPng(progress, mdLines)
		mdLines = replaceLineBreaks(progress, mdLines)
		mdLines = correctTableSeparators(progress, mdLines)
		writeMDFile(progress, mdLines, document, outDirectory)


def main(args=None):
	# Parse command line arguments
	parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
	parser.add_argument('--outdir', '-o', action='store', dest='outDirectory', default = 'out', metavar = '<output directory>',  help = 'specify output directory')
	parser.add_argument('document',  help = 'document to parse')
	args = parser.parse_args()

	# Process documents and print output
	os.makedirs(args.outDirectory, exist_ok = True)

	process(args.document, args.outDirectory)

if __name__ == '__main__':
    sys.exit(main())