#
#	publish_specs.sh
#
#	Script to generate documents from markdown files with Pandoc. Outputs:
#		- .docx
#		- .pdf
#		- .epub 
#
#	(c) 2024 by Miguel Angel Reina Ortega
#	License: BSD 3-Clause License. See the LICENSE file for further details.
#
#!/bin/bash

DOCKER_IMAGE=pandoc:master
TOOLS_DOCKER_IMAGE=generatechangemarks:master

echo "\n------ Checking for docker image --------"
docker pull "$DOCKER_IMAGE"

echo "------ Removing previous outputs --------"
rm **/*.docx

echo "------ Parsing repo URL --------"

HOST_URL=$(echo $1 | cut -d'/' -f 1-3)
PROJECT_NAME=$(echo $1 | cut -d'/' -f 5- | cut -d'.' -f 1)
echo "HOST URL:" $HOST_URL
echo "PROJECT NAME:" $PROJECT_NAME
echo "TAG NAME:" $2
echo "SPEC NAME:" $4

echo "------ Getting .md file(s) ------"
# If there are no .md files, then simply exit
ls | grep -q 'md'
specs=$?
if [ ! $specs ] ; then 
    echo "-- No Markdown files."
    exit 0
fi

for i in $(find -name "*.svg") ; do
    png="${i%.svg}.png"
    if [ -f "$png" ]; then
      echo "Skipping $i — PNG already exists at $png"
      continue
    fi
	echo "\n------ Converting SVG to PNG for pandoc --------"
	docker run --rm -v $(pwd):/tmp/ -u $(id -u):$(id -g) "$TOOLS_DOCKER_IMAGE" svg2png "/tmp/$i"
done

for i in *.md ;  do
    if [ $i != 'README.md' ]; then
    	if [[ $i =~ (TS|TR).*\.md ]] ; then
        	echo "\n------ Adding TOC to spec --------"
	      	docker run --rm -v $(pwd):/tmp/ -u $(id -u):$(id -g) "$TOOLS_DOCKER_IMAGE" generateTOC --add-content "/tmp/$i"
	  	fi
	    echo "\n------ Preparaing spec --------"
	    docker run --rm -v $(pwd):/tmp/ -u $(id -u):$(id -g) "$TOOLS_DOCKER_IMAGE" pandocFilter -o "/tmp/" "/tmp/$i"
	    echo "\n------ Publishing spec --------"
		mkdir baseline
	    # Word output
	    docker run --rm -v $(pwd):/data -u $(id -u):$(id -g) "$DOCKER_IMAGE" "/data/$i" -F mermaid-filter -f markdown+escaped_line_breaks -t docx --reference-doc "/data/${3}" -o "baseline/${4}_${2}.docx"
	    #echo 'docker run --rm -v $(pwd):/data -u $(id -u):$(id -g) "$DOCKER_IMAGE" "/data/$i" -F mermaid-filter -f markdown+escaped_line_breaks -t pdf -o "${4}_${2}.pdf"'
	    if [ "$5" != "onlyDocx" ]; then
			# PDF Output
			docker run --rm -v $(pwd):/data -u $(id -u):$(id -g) "$DOCKER_IMAGE" "/data/$i" -F mermaid-filter -f markdown+escaped_line_breaks -t pdf -o "baseline/${4}_${2}.pdf"
			# EPUB Output
			docker run --rm -v $(pwd):/data -u $(id -u):$(id -g) "$DOCKER_IMAGE" "/data/$i" -f markdown+escaped_line_breaks -t epub3 -o "baseline/${4}_${2}.epub" --metadata title="${PROJECT_NAME}_${2}" --metadata creator="oneM2M Partnership Project" --metadata rights="Copyright 2024 oneM2M Partners
hip Project"
		fi
	fi
done


exit 0
