#
#	toMkdocs.py
#
#	(c) 2024 by Andreas Kraft & Miguel Angel Reina Ortega
#	License: BSD 3-Clause License. See the LICENSE file for further details.

#
#	This script converts oneM2M spec markdown file to a mkdocs compatible
#	directory structure.
#
from __future__ import annotations

import argparse, os, shutil
from rich import print
from markdownTools import Line, Document, analyseMarkdown, setLoggers
from regexMatches import match2spaceListIndention

verbose = False
veryVerbose = False


def printDebug(text:str) -> None:
	"""	Print a debug message.

		Args:
			text: The text of the debug message.
	"""
	if verbose:
		print(f'[dim]{text}')


def printInfo(text:str) -> None:
	"""	Print an information message.

		Args:
			text: The text of the information message.
	"""
	print(f'[green]{text}')


def printWarning(text:str) -> None:
	"""	Print a warning message.

		Args:
			text: The text of the warning message.
	"""
	print(f'[yellow]{text}')


def printError(text:str) -> None:
	"""	Print an error message.

		Args:
			text: The text of the error message.
	"""
	print(f'[red]{text}')

def prepareForMkdocs(document:Document, includeHangingParagraphs:bool = False) -> None:
	"""	Prepare the clauses for MkDocs. This includes removing the heading
		from the clauses and marking the clauses that are only for navigation.

		After the preparation, the clauses are stored in the document object.

		Args:
			document: The document object.
			includeHangingParagraphs: Include hanging paragraphs in the output.
	"""

	# Remove the heading from the lines. The heading is the first line
	# in the clause. This is done because MkDocs repeats the heading when
	# displaying the page.
	for clause in document.clauses:
		if clause.linesCount > 0:
			clause.lines.pop(0)
			# Also, remove the first empty lines if they exist
			while clause.linesCount > 0 and clause.lines[0].text.strip() == '':
				clause.lines.pop(0)

		# Detect and handle hanging paragraphs. This is extra text in a clause, which
	# has sub-clauses. This text is not allowed in oneM2M specifications.
	for i, clause in enumerate(document.clauses):
		if clause.level > 0 and clause.linesCount > 0:
			# Check if there is a sub-clause in the next clause
			if i + 1 < len(document.clauses) and document.clauses[i+1].level > clause.level:
				# This is a hanging paragraph. Remove the text from the current clause.
				printWarning(f'Hanging paragraph in clause "{clause.title}" {"(removed)" if not includeHangingParagraphs else "(kept)"}')
				if not includeHangingParagraphs:
					document.clauses[i].lines = []
				else:
					# Add a note to the hanging paragraph
					document.clauses[i].lines = [Line("<mark>Editor note: This is a hanging paragraph and it must be moved to its own clause</mark>")] + [Line()] + document.clauses[i].lines

	# Repair wrong markdown for indented lines.
	# Add 2 spaces to existing 2-space indentions
	for clause in document.clauses:
		for i, line in enumerate(clause.lines):
			if match2spaceListIndention.match(line.text):
				clause.lines[i].text = '  ' + line.text


def writeClausesMkDocs(document:Document, filename:str, navTitle:str, addNavTitle:bool = False) -> None:
	"""	Write the clauses to separate files and create a navigation file.

		Args:
			document: The document object.
			filename: The name of the original markdown file.
			navTitle: The title of the navigation entry. This is used to determine the directories.
			addNavTitle: Add the title as an extra navigation level to the navigation file.
	"""

	printInfo(f'Writing clauses to files')
	# create directory first
	os.makedirs(f'{os.path.dirname(filename)}/{navTitle}', exist_ok = True)

	# Write the files
	for i, f in enumerate(document.clauses):
		# write to single files, even empty ones
		printDebug(f'Writing "{f.clauseNumber}.md" - "{f.title}"')
		with open(f'{os.path.dirname(filename)}/{navTitle}/{f.clauseNumber}.md', 'w') as file:
			# Add one empty line before the clause. This is done to avoid
			# a bug in MkDocs that does not display the first line of a clause
			# if it contains a colon. It does not matter otherwise if the line
			# is empty or not.
			file.writelines(f.asStringList(1))	

	
	# write nav.yml file
	printInfo(f'Writing "_nav.yml"')
	indentation = '  ' if addNavTitle else ''	# TODO make number of spaces configurable
	with open(f'{os.path.dirname(filename)}/_nav.yml', 'w') as file:
		printDebug(f'Writing navigation file')
		if addNavTitle:
			file.write(f'{indentation}- {navTitle}:\n')
		for i, f in enumerate(document.clauses):

			if not f.title:
				# print("continue")
				continue

			# TODO handle if the next clause is more than one level deeper

			_title = f.title.replace("'", '"')
			nextClause = document.clauses[i+1] if i+1 < len(document.clauses) else None
			if nextClause is None or nextClause.level <= f.level:
				file.write(f"{indentation}{'  '*f.level}- '{_title}': '{navTitle}/{f.clauseNumber}.md'\n")
			else:
				file.write(f"{indentation}{'  '*f.level}- '{_title}':\n")
				if len(f) > 0:
					file.write(f"{indentation}{'  '*nextClause.level}- 'Hanging paragraph': '{navTitle}/{f.clauseNumber}.md'\n")


# TODO handle multiple nav levels (left bar) better (make conifgurable)


def copyMediaFiles(filename:str, navTitle:str, mediaDirectory:str = 'media') -> None:
	"""	Copy media files from the source directory to the target directory.

		Args:
			filename: The name of the markdown file.
			navTitle: The title of the navigation entry.
			mediaDirectory: The name of the media directory.
	"""
	sourceDirectory = f'{os.path.dirname(filename)}/{mediaDirectory}'
	targetDirectory = f'{os.path.dirname(filename)}/{navTitle}/{mediaDirectory}'

	if os.path.exists(sourceDirectory):
		printInfo(f'Copying media files from "{sourceDirectory}" to "{targetDirectory}"')
		shutil.copytree(sourceDirectory, targetDirectory, dirs_exist_ok = True)
	else:
		printError(f'Media directory "{sourceDirectory}" does not exist')

	
def processDocument(args:argparse.Namespace) -> None:
	global verbose, veryVerbose
	inDocumentFilename = os.path.abspath(args.document)
	veryVerbose = args.very_verbose
	verbose = args.verbose
	if veryVerbose:
		verbose = True

	# Analyse the markdown file
	document = analyseMarkdown(inDocumentFilename)
	document.splitMarkdownDocument(args.ignore_clause, args.split_level)
	document.insertFootnotes()
	document.updateLinks()
	document.updateNotes()
	
	prepareForMkdocs(document, args.include_hanging_paragraphs)

	# Write the clauses to files
	writeClausesMkDocs(document, inDocumentFilename, args.title, args.nav_add_title)

	# Copy the media files
	copyMediaFiles(inDocumentFilename, args.title, args.media_directory)


def main() -> None:
	parser = argparse.ArgumentParser(description = 'Convert oneM2M markdown specificatios to MkDocs format',
									 formatter_class = argparse.ArgumentDefaultsHelpFormatter)

	parser.add_argument('--verbose', '-v', action = 'store_true', help = 'verbose output during processing')
	parser.add_argument('--out', '-o', metavar='outfile', help = 'write output to file instead of stdout')
	parser.add_argument('--very-verbose', '-vv', action = 'store_true', help = 'very verbose output during processing')
	parser.add_argument('--ignore-clause', '-ic', metavar = 'clause', nargs = '+', default = [ 'Contents', 'History' ], help = 'ignore headers in the markdown document')
	parser.add_argument('--include-hanging-paragraphs', '-ihp', action = 'store_true', default = False, help = 'include hanging paragraphs (text in clauses with sub-clauses) in the output files')
	parser.add_argument('--include-title', '-it', action = 'store_true', help = 'include the content before the first heading in the output files as "0.md"')
	parser.add_argument('--split-level', '-sl', metavar = 'level', type = int, default = 2, help = 'on which level to split clauses to separate files')
	parser.add_argument('--media-directory', '-md', metavar = 'media-directory', default = 'media', help = 'directory name where media files are stored')
	parser.add_argument('--title', '-t', metavar = 'title', required = True, help = 'mkdocs navigation tile')
	parser.add_argument('--nav-add-title', '-nat', action = 'store_true', default = False, help = 'add the title as an extra navigation level to the navigation file')

	parser.add_argument('document', type = str, help = 'a oneM2M markdown specification document to process')
	args = parser.parse_args()
	setLoggers(info = printInfo, 
			   debug = printDebug,
			   error = printError)
	processDocument(args)


if __name__ == '__main__':
	main()

	
