#
#	updateIndex.py
#
#	Script to update index.htm file on Pages.
#
#	(c) 2023 by Miguel Angel Reina Ortega
#	License: BSD 3-Clause License. See the LICENSE file for further details.
#
from bs4 import BeautifulSoup
import argparse, os


def upgradeIndex(args: argparse.Namespace) -> None:
    # Check if index.htm exists
    if not os.path.exists('public/baselines/index.html'):
        print("'index.html' does not exist yet, Upgrade is not needed")
        exit()
    else:
        # Read the source HTML file
        with open('public/baselines/index.html', 'r', encoding='utf-8') as source_file:
            source_html = source_file.read()

        # Read the destination HTML file
        with open(args.indexTemplate, 'r', encoding='utf-8') as dest_file:
            dest_html = dest_file.read()

        print(source_html)
        print(dest_html)
        # Parse both HTML files
        source_soup = BeautifulSoup(source_html, 'html.parser')
        dest_soup = BeautifulSoup(dest_html, 'html.parser')

        print(source_soup)
        print(dest_soup)
        # Find the <ul> element in both the source and destination HTML
        source_ul = source_soup.find('ul', id='document-list')
        dest_ul = dest_soup.find('ul', id='document-list')

        print(source_ul)
        print(dest_ul)
        # Copy the contents of the source <ul> to the destination <ul>
        if dest_ul and source_ul:
            dest_ul.contents.extend(source_ul.contents)
        else:
            print("ERROR: List not found in source and destination index files...")
            exit(1)

        # Save the updated destination HTML
        with open('public/baselines/index.html', 'w', encoding='utf-8') as dest_updated_file:
            dest_updated_file.write(str(dest_soup))

        print("Index.html upgraded")


def updateIndex(args:argparse.Namespace) -> None:
    
    # Check if index.htm exists
    if not os.path.exists('public/baselines/index.html'):
        print("'index.html' does not exist yet, using template")
        with open(args.indexTemplate, 'r', encoding='utf-8') as file:
            html_content = file.read()
    else:
        # Read index HTML
        with open('public/baselines/index.html', 'r', encoding='utf-8') as file:
            html_content = file.read()

    # Analyze index content with BeautifulSoup
    soup = BeautifulSoup(html_content, 'html.parser')

    if args.removeEntry:
        # Remove the element <li> for the document
        if args.documentLink is not None and args.documentName is not None:
            # Find the element to delete
            element_to_delete = soup.find('li', string=args.documentName)
            # Check if the element is found before attempting to delete it
            if element_to_delete:
                # Remove the element from the HTML tree
                element_to_delete.decompose()
            else:
                print("Element not found.")
    elif soup.find('li', string=args.documentName):
        print("'index.html' does already contain the entry")
    else:
        # Find document list
        # Assume there's a list (<ul>) with id "document-list"
        document_list = soup.find('ul', id='document-list')
        # Add the new element <li> for the new document
        if args.documentLink is not None and args.documentName is not None:
            li = soup.new_tag('li')
            a = soup.new_tag('a', href=args.documentLink)
            a.string = args.documentName
            li.append(a)
            document_list.append(li)

    # Save the updated index.html file
    with open('public/baselines/index.html', 'w', encoding='utf-8') as file:
        file.write(str(soup))

    print("Index.html updated")
			


if __name__ == '__main__':

    # Parse command line arguments
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-re', '--removeEntry', action='store_true', required=False, default=False,
                        help="Indicate to remove the given entry")
    parser.add_argument('-u', '--upgrade', action='store_true', required=False, default=False,
                        help="Upgrade the index.html")
    parser.add_argument('documentName', help = 'Document name to add to the list')
    parser.add_argument('documentLink', help = 'Document location to add to the list')
    parser.add_argument('indexTemplate', help = 'Index template document')
    args = parser.parse_args()

    if args.upgrade:
        upgradeIndex(args)
    else:
        updateIndex(args)

